@@ -2,24 +2,24 @@
  * Gnome Chemistry Utils
  * element.cc 
  *
- * Copyright (C) 2002-2005
+ * Copyright (C) 2002-2006
  *
  * Developed by Jean Bréfort <jean.brefort@normalesup.org>
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
+ * This program is free software; you can redistribute it and/or 
+ * modify it under the terms of the GNU General Public License as 
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
  *
- * This library is distributed in the hope that it will be useful,
+ * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the 
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA  02111-1307, USA.
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
+ * USA
  */
 
 #include "config.h"
@@ -31,22 +31,27 @@
 #include <locale.h>
 #include <vector>
 #include <map>
+#include <set>
 #include <string>
 #include <libintl.h>
+#include <math.h>
 #include <string.h>
-#define _(String) gettext(String)
+#include <glib/gi18n-lib.h>
+#include <goffice/utils/go-math.h>
+
+static set<string>units;
 
 static void ReadValue (char const *source, GcuValue &value)
 {
-	char *tmp, *dot;
-	value.value = strtod (source, &tmp);
+	char *buf, *dot;
+	value.value = strtod (source, &buf);
 	dot = strchr (source, '.');
-	value.prec = (dot)? tmp - dot - 1: 0;
-	value.delta = (*tmp == '(')? strtol (tmp + 1, NULL, 10): 0;
+	value.prec = (dot)? buf - dot - 1: 0;
+	value.delta = (*buf == '(')? strtol (buf + 1, NULL, 10): 0;
 }
 
-using namespace gcu;
-using namespace std;
+namespace gcu
+{
 
 class EltTable
 {
@@ -64,6 +69,10 @@ private:
 	map <string, Element*> EltsMap;
 };
 
+} // namespace gcu
+
+using namespace gcu;
+
 EltTable Table;
 
 EltTable::EltTable()
@@ -75,8 +84,15 @@ EltTable::EltTable()
 	xmlDocPtr xml;
 	char* DefaultName;
 	char *lang = getenv ("LANG");
-	char *old_num_locale, *tmp, *num, *dot;
+	setlocale (LC_ALL, lang);
+	char *old_num_locale, *buf, *num, *dot;
 	unsigned char Z;
+	map <string, string> Langs;
+	Langs["de"] = _("German");
+	Langs["fr"] = _("French");
+	Langs["it"] = _("Italian");
+	Langs["pl"] = _("Polish");
+	Langs["ru"] = _("Russian");
 	if (!(xml = xmlParseFile (DATADIR"/gchemutils-unstable/elements.xml")))
 	{
 		g_error (_("Can't find and read elements.xml"));
@@ -84,26 +100,26 @@ EltTable::EltTable()
 	old_num_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
 	setlocale (LC_NUMERIC, "C");
 	xmlNode* node = xml->children, *child;
-	if (strcmp ((const char*) node->name, "gpdata")) g_error (_("Uncorrect file format: elements.xml"));
+	if (strcmp ((const char*) node->name, "gpdata")) g_error (_("Incorrect file format: elements.xml"));
 	node = node->children;
 	Element* Elt;
 	while (node)
 	{
 		if (strcmp ((const char*) node->name, "text"))
 		{
-			if (strcmp ((const char*) node->name, "element")) g_error (_("Uncorrect file format: elements.xml"));
-			tmp = (char*) xmlGetProp (node, (xmlChar*) "symbol");
+			if (strcmp ((const char*) node->name, "element")) g_error (_("Incorrect file format: elements.xml"));
+			buf = (char*) xmlGetProp (node, (xmlChar*) "symbol");
 			num = (char*) xmlGetProp (node, (xmlChar*) "Z");
-			Elt = new Element (Z = atoi (num), tmp);
+			Elt = new Element (Z = atoi (num), buf);
 			xmlFree (num);
-			xmlFree (tmp);
+			xmlFree (buf);
 			num = (char*) xmlGetProp (node, (xmlChar*) "max_bonds");
 			Elt->m_MaxBonds = atoi (num);
 			xmlFree (num);
 			num = (char*) xmlGetProp (node, (xmlChar*) "weight");
-			Elt->m_Weight = strtod (num, &tmp);
+			Elt->m_Weight = strtod (num, &buf);
 			dot = strchr (num, '.');
-			Elt->m_WeightPrec = (dot)? tmp - dot - 1: 0;
+			Elt->m_WeightPrec = (dot)? buf - dot - 1: 0;
 			child = node->children;
 			DefaultName = NULL;
 			while (child)
@@ -113,31 +129,35 @@ EltTable::EltTable()
 					continue;
 				}
 				if (!strcmp((const char*)child->name, "name")) {
-					tmp = (char*) xmlNodeGetLang (child);
-					if ((tmp) && (lang) && (!strncmp (lang, tmp, 2))) {
-						xmlFree (tmp);
-						tmp = (char*) xmlNodeGetContent (child);
-						Elt->name = tmp;
-						xmlFree (tmp);
-					} else if (!tmp)
+					buf = (char*) xmlNodeGetLang (child);
+					if ((buf) && (lang)){
+						string Lang = Langs[buf];
+						char *Name = (char*) xmlNodeGetContent (child);
+						if (Lang.length ())
+							Elt->names[Lang] = Name;
+						if (!strncmp (lang, buf, 2))
+							Elt->name = Name;
+						xmlFree (Name);
+					} else if (!buf) {
 						DefaultName = (char*) xmlNodeGetContent (child);
-					else
-						xmlFree (tmp);
+						Elt->names[_("English")] = DefaultName;
+					}
+					xmlFree (buf);
 				} else if (!strcmp ((const char*) child->name, "color")) {
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "red");
-					if (tmp) {
-						Elt->m_DefaultColor[0] = strtod (tmp, NULL);
-						xmlFree (tmp);
+					buf = (char*) xmlGetProp (child, (xmlChar*) "red");
+					if (buf) {
+						Elt->m_DefaultColor[0] = strtod (buf, NULL);
+						xmlFree (buf);
 					}
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "green");
-					if (tmp) {
-						Elt->m_DefaultColor[1] = strtod (tmp, NULL);
-						xmlFree (tmp);
+					buf = (char*) xmlGetProp (child, (xmlChar*) "green");
+					if (buf) {
+						Elt->m_DefaultColor[1] = strtod (buf, NULL);
+						xmlFree (buf);
 					}
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "blue");
-					if (tmp) {
-						Elt->m_DefaultColor[2] = strtod (tmp, NULL);
-						xmlFree (tmp);
+					buf = (char*) xmlGetProp (child, (xmlChar*) "blue");
+					if (buf) {
+						Elt->m_DefaultColor[2] = strtod (buf, NULL);
+						xmlFree (buf);
 					}
 				}				child = child->next;
 			}
@@ -336,7 +356,8 @@ unsigned Element::GetMaxBonds(gint Z)
 bool Element::GetRadius(GcuAtomicRadius* radius)
 {
 	Element* Elt = Table[radius->Z];
-	if (!Elt) return false;
+	if (!Elt || !Elt->m_radii.size ())
+		return false;
 	for (int i = 0; Elt->m_radii[i]; i++)
 	{
 		if (radius->type != Elt->m_radii[i]->type) continue;
@@ -395,7 +416,7 @@ double Element::GetWeight (int Z, int &prec)
 void Element::LoadRadii ()
 {
 	xmlDocPtr xml;
-	char *old_num_locale, *tmp, *num, *dot, *end;
+	char *old_num_locale, *buf, *num;
 	unsigned char Z;
 	static bool loaded = false;
 	if (loaded)
@@ -407,13 +428,18 @@ void Element::LoadRadii ()
 	old_num_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
 	setlocale (LC_NUMERIC, "C");
 	xmlNode* node = xml->children, *child;
-	if (strcmp ((const char*) node->name, "gpdata")) g_error (_("Uncorrect file format: radii.xml"));
+	if (strcmp ((const char*) node->name, "gpdata")) g_error (_("Incorrect file format: radii.xml"));
 	node = node->children;
 	Element* Elt;
+	set<string>::iterator it = units.find ("pm");
+	if (it == units.end ()) {
+		units.insert ("pm");
+		it = units.find ("pm");
+	}
 	while (node) {
 		if (strcmp ((const char*) node->name, "text"))
 		{
-			if (strcmp ((const char*) node->name, "element")) g_error (_("Uncorrect file format: radii.xml"));
+			if (strcmp ((const char*) node->name, "element")) g_error (_("Incorrect file format: radii.xml"));
 			num = (char*) xmlGetProp (node, (xmlChar*) "Z");
 			Elt = Table[Z = atoi (num)];
 			child = node->children;
@@ -426,51 +452,50 @@ void Element::LoadRadii ()
 				if (!strcmp ((const char*) child->name, "radius")) {
 					GcuAtomicRadius* radius = new GcuAtomicRadius;
 					radius->Z = Z;	//FIXME: is it really useful there?
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "type");
-					if (!tmp ||
-						((!((!strcmp (tmp, "covalent")) && (radius->type = GCU_COVALENT))) &&
-						(!((!strcmp (tmp, "vdW")) && (radius->type = GCU_VAN_DER_WAALS))) &&
-						(!((!strcmp (tmp, "ionic")) && (radius->type = GCU_IONIC))) &&
-						(!((!strcmp (tmp, "metallic")) && (radius->type = GCU_METALLIC))) &&
-						(!((!strcmp (tmp, "atomic")) && ((radius->type = GCU_ATOMIC) || true))))) {
+					buf = (char*) xmlGetProp (child, (xmlChar*) "type");
+					if (!buf ||
+						((!((!strcmp (buf, "covalent")) && (radius->type = GCU_COVALENT))) &&
+						(!((!strcmp (buf, "vdW")) && (radius->type = GCU_VAN_DER_WAALS))) &&
+						(!((!strcmp (buf, "ionic")) && (radius->type = GCU_IONIC))) &&
+						(!((!strcmp (buf, "metallic")) && (radius->type = GCU_METALLIC))) &&
+						(!((!strcmp (buf, "atomic")) && ((radius->type = GCU_ATOMIC) || true))))) {
 						//invalid radius
 						delete radius;
-						if (tmp)
-							xmlFree (tmp);
+						if (buf)
+							xmlFree (buf);
 						continue;
 					}
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "scale");
-					if (tmp) {
-						radius->scale = g_strdup (tmp);
-						xmlFree (tmp);
+					buf = (char*) xmlGetProp (child, (xmlChar*) "scale");
+					if (buf) {
+						radius->scale = g_strdup (buf);
+						xmlFree (buf);
 					} else
 						radius->scale = NULL;
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "charge");
-					if (tmp) {
-						radius->charge = strtol (tmp, NULL, 10);
-						xmlFree (tmp);
+					buf = (char*) xmlGetProp (child, (xmlChar*) "charge");
+					if (buf) {
+						radius->charge = strtol (buf, NULL, 10);
+						xmlFree (buf);
 					} else
 						radius->charge = 0;
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "cn");
-					if (tmp) {
-						radius->cn = strtol (tmp, NULL, 10);
-						xmlFree (tmp);
+					buf = (char*) xmlGetProp (child, (xmlChar*) "cn");
+					if (buf) {
+						radius->cn = strtol (buf, NULL, 10);
+						xmlFree (buf);
 					} else
 						radius->cn = -1;
-					tmp = (char*) xmlGetProp (child, (xmlChar*)"spin");
-					if ((!tmp) ||
-						(!((!strcmp (tmp, "low")) && (radius->spin = GCU_LOW_SPIN))) &&
-						(!((!strcmp (tmp, "high")) && (radius->spin = GCU_HIGH_SPIN))))
+					buf = (char*) xmlGetProp (child, (xmlChar*)"spin");
+					if ((!buf) ||
+						(!((!strcmp (buf, "low")) && (radius->spin = GCU_LOW_SPIN))) &&
+						(!((!strcmp (buf, "high")) && (radius->spin = GCU_HIGH_SPIN))))
 						radius->spin = GCU_N_A_SPIN;
-					if (tmp)
-						xmlFree (tmp);
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "value");
-					if (tmp) {
-						radius->value.value = strtod (tmp, &end);
-						dot = strchr (tmp, '.');
-						radius->value.prec = (dot)? end - dot - 1: 0;
+					if (buf)
+						xmlFree (buf);
+					buf = (char*) xmlGetProp (child, (xmlChar*) "value");
+					if (buf) {
+						ReadValue (buf, (GcuValue&) radius->value) ;
+						radius->value.unit = (*it).c_str ();
 						Elt->m_radii.push_back (radius);
-						xmlFree (tmp);
+						xmlFree (buf);
 					} else
 						delete radius;
 				} else
@@ -490,8 +515,9 @@ void Element::LoadRadii ()
 void Element::LoadElectronicProps ()
 {
 	xmlDocPtr xml;
-	char *old_num_locale, *tmp, *num, *dot, *end;
+	char *old_num_locale, *buf, *num, *dot, *end;
 	unsigned char Z;
+	unsigned i;
 	static bool loaded = false;
 	if (loaded)
 		return;
@@ -502,12 +528,12 @@ void Element::LoadElectronicProps ()
 	old_num_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
 	setlocale (LC_NUMERIC, "C");
 	xmlNode* node = xml->children, *child;
-	if (strcmp ((const char*) node->name, "gpdata")) g_error (_("Uncorrect file format: elecprops.xml"));
+	if (strcmp ((const char*) node->name, "gpdata")) g_error (_("Incorrect file format: elecprops.xml"));
 	node = node->children;
 	Element* Elt;
 	while (node) {
 		if (strcmp ((const char*) node->name, "text")) {
-			if (strcmp ((const char*) node->name, "element")) g_error (_("Uncorrect file format: elecprops.xml"));
+			if (strcmp ((const char*) node->name, "element")) g_error (_("Incorrect file format: elecprops.xml"));
 			num = (char*) xmlGetProp (node, (xmlChar*) "Z");
 			Elt = Table[Z = atoi (num)];
 			child = node->children;
@@ -516,24 +542,121 @@ void Element::LoadElectronicProps ()
 					child = child->next;
 					continue;
 				}
-				if (!strcmp ((const char*) child->name, "electronegativity")) {
+				if (!strcmp ((const char*) child->name, "en")) {
 					GcuElectronegativity* en = new GcuElectronegativity;
 					en->Z = Z;	//FIXME: is it really useful there?
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "scale");
-					if (tmp) {
-						en->scale = g_strdup (tmp);
-						xmlFree (tmp);
+					buf = (char*) xmlGetProp (child, (xmlChar*) "scale");
+					if (buf) {
+						en->scale = g_strdup (buf);
+						xmlFree (buf);
 					} else
 						en->scale = NULL;
-					tmp = (char*) xmlGetProp (child, (xmlChar*) "value");
-					if (tmp) {
-						en->value.value = strtod (tmp, &end);
-						dot = strchr (tmp, '.');
+					buf = (char*) xmlGetProp (child, (xmlChar*) "value");
+					if (buf) {
+						en->value.value = strtod (buf, &end);
+						dot = strchr (buf, '.');
 						en->value.prec = (dot)? end - dot - 1: 0;
+						en->value.delta = 0; // we should use a generic parser
 						Elt->m_en.push_back (en);
-						xmlFree (tmp);
+						xmlFree (buf);
 					} else
 						delete en;	//without a value, the structure is useless and is discarded
+				} else if (!strcmp ((const char*) child->name, "config")) {
+					buf = (char*) xmlNodeGetContent (child);
+					char *cur = buf;
+					bool nonvoid = false;
+					if (buf[0] == '[') {
+						Elt->ElecConfig.append (buf, 4);
+						cur += 4;
+						nonvoid = true;
+					}
+					while (cur && *cur) {
+						if (nonvoid) {
+							cur++;
+							Elt->ElecConfig.append (" ");
+						}
+						Elt->ElecConfig.append (cur, 2);
+						cur += 2;
+						i = 1;
+						while (cur[i] > ' ')
+							i++;
+						Elt->ElecConfig.append ("<sup>");
+						Elt->ElecConfig.append (cur, i);
+						Elt->ElecConfig.append ("</sup>");
+						cur += i;
+					}
+					Elt->ElecConfig.append (" ");
+					xmlFree (buf);
+				} else if (!strcmp ((const char*) child->name, "ei")) {
+					unsigned rank;
+					buf = (char*) xmlGetProp (child, (xmlChar*) "rank");
+					if (buf) {
+						rank = strtol (buf, NULL, 10);
+						xmlFree (buf);
+					} else
+						rank = 1;
+					if ((i = Elt->m_ei.size ()) < rank) {
+						Elt->m_ei.resize (rank);
+						for (; i < rank; i++)
+							Elt->m_ei[i].value = go_nan;
+					}
+					rank--;
+					buf = (char*) xmlGetProp (child, (xmlChar*) "value");
+					if (buf) {
+						ReadValue (buf, (GcuValue&)Elt->m_ei[rank]) ;
+						xmlFree (buf);
+					} else {
+						//no need to read the unit
+						Elt->m_ei[rank].value = go_nan;
+						break;
+					}
+					buf = (char*) xmlGetProp (child, (xmlChar*) "unit");
+					if (buf) {
+						string str(buf);
+						set<string>::iterator it = units.find (str);
+						if (it == units.end ()) {
+							units.insert (str);
+							it = units.find (str);
+						}
+						Elt->m_ei[rank].unit = (*it).c_str ();
+						xmlFree (buf);
+					} else
+						Elt->m_ei[rank].unit = "MJ.mol<sup>-1</sup>";
+				} else if (!strcmp ((const char*) child->name, "ae")) {
+					unsigned rank;
+					buf = (char*) xmlGetProp (child, (xmlChar*) "rank");
+					if (buf) {
+						rank = strtol (buf, NULL, 10);
+						xmlFree (buf);
+					} else
+						rank = 1;
+					if ((i = Elt->m_ae.size ()) < rank) {
+						Elt->m_ae.resize (rank);
+						for (; i < rank; i++)
+							Elt->m_ae[i].value = go_nan;
+					}
+					rank--;
+					buf = (char*) xmlGetProp (child, (xmlChar*) "value");
+					if (buf) {
+						ReadValue (buf, (GcuValue&)Elt->m_ae[rank]) ;
+						xmlFree (buf);
+					} else {
+						//no need to read the unit
+						Elt->m_ae[rank].value = go_nan;
+						break;
+					}
+					buf = (char*) xmlGetProp (child, (xmlChar*) "unit");
+					if (buf) {
+						string str(buf);
+						set<string>::iterator it = units.find (str);
+						if (it == units.end ()) {
+							units.insert (str);
+							it = units.find (str);
+						}
+						Elt->m_ae[rank].unit = (*it).c_str ();
+						xmlFree (buf);
+					} else
+						Elt->m_ae[rank].unit = "kJ.mol<sup>-1</sup>";
 				} else
 					g_error ("Invalid property node");
 				child = child->next;
@@ -551,7 +674,7 @@ void Element::LoadElectronicProps ()
 void Element::LoadIsotopes ()
 {
 	xmlDocPtr xml;
-	char *old_num_locale, *tmp, *num;
+	char *old_num_locale, *num;
 	unsigned char Z;
 	static bool loaded = false;
 	if (loaded)
@@ -563,14 +686,14 @@ void Element::LoadIsotopes ()
 	old_num_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
 	setlocale (LC_NUMERIC, "C");
 	xmlNode* node = xml->children, *child;
-	if (strcmp ((const char*) node->name, "gpdata")) g_error (_("Uncorrect file format: isotopes.xml"));
+	if (strcmp ((const char*) node->name, "gpdata")) g_error (_("Incorrect file format: isotopes.xml"));
 	node = node->children;
 	Element *Elt;
 	Isotope *Is;
 	int minA, maxA, niso;
 	while (node) {
 		if (strcmp ((const char*) node->name, "text")) {
-			if (strcmp ((const char*) node->name, "element")) g_error (_("Uncorrect file format: isotopes.xml"));
+			if (strcmp ((const char*) node->name, "element")) g_error (_("Incorrect file format: isotopes.xml"));
 			minA = maxA = niso = 0;
 			num = (char*) xmlGetProp (node, (xmlChar*) "Z");
 			Elt = Table[Z = atoi (num)];
@@ -651,7 +774,7 @@ IsotopicPattern *Element::GetIsotopicPattern (unsigned natoms)
 	IsotopicPattern *pat, *pattern, *result = NULL;
 	if (natoms == 0)
 		return NULL;
-	int i = 1;
+	unsigned i = 1;
 	while ((natoms & 1) == 0) {
 		natoms >>= 1;
 		i++;
@@ -662,14 +785,14 @@ IsotopicPattern *Element::GetIsotopicPattern (unsigned natoms)
 			result->Ref ();
 		} else if (natoms & 1) {
 			while (m_patterns.size () < i) {
-				pat = m_patterns[m_patterns.size () - 1]->square ();
+				pat = m_patterns[m_patterns.size () - 1]->Square ();
 				pattern = pat->Simplify ();
 				pat->Unref ();
 				m_patterns.push_back (pattern);
 			}
 			pattern = m_patterns[i - 1];
 			if (result) {
-				pat = result->multiply (*pattern);
+				pat = result->Multiply (*pattern);
 				result->Unref ();
 				result = pat->Simplify ();
 				pat->Unref ();
@@ -683,3 +806,13 @@ IsotopicPattern *Element::GetIsotopicPattern (unsigned natoms)
 	}
 	return result;
 }
+
+GcuDimensionalValue const *Element::GetIonizationEnergy (unsigned rank)
+{
+	return (rank <= m_ei.size ())? &m_ei[rank - 1]: NULL;
+}
+
+GcuDimensionalValue const *Element::GetElectronAffinity (unsigned rank)
+{
+	return (rank <= m_ae.size ())? &m_ae[rank - 1]: NULL;
+}
